#include <stdlib.h>
#include <inttypes.h>

#include "pagetable.h"

#define LEVEL_SIZE 1024
#define LEVEL_MASK 0x3FF
#define LEVEL1_BIT 22
#define LEVEL2_BIT 12

typedef struct level {
    unsigned int count;
    void *childs[LEVEL_SIZE];
} level_t;

static level_t level1;

/* Initialiser et niveau i sidetabellen */
static void init_level(level_t *l) {
    int i=0;

    l->count = 0;

    for(i=0; i<LEVEL_SIZE; i++) {
        l->childs[i] = NULL;
    }
}

/* Initialiser sidetabellen */
void pagetable_init() {
    init_level(&level1);
}

/*
Indstter side informationen fra 'page' p pladsen for den logiske side
'logical'. Det antages at funktionen ikke kan fejle.
*/
void pagetable_set(uint32_t logical, page_entry_t *page) {
    page_entry_t *pe;
    level_t *l2;
    uint32_t p1, p2;

    p1 = (logical >> LEVEL1_BIT) & LEVEL_MASK;

    /* Opslag p frste niveau. */
    l2 = level1.childs[p1];

    /* Skal et nyt niveau allokeres.*/
    if (l2 == NULL) {
        l2 = level1.childs[p1] = malloc(sizeof(level_t));
        init_level(l2);
        level1.count++;
    }

    p2 = (logical >> LEVEL2_BIT) & LEVEL_MASK;

    /* Opslag p andet niveau. */
    pe = l2->childs[p2];

    /* Skal der allokeres et nyt sideelement. */
    if (pe == NULL) {
        pe = l2->childs[p2] = malloc(sizeof(page_entry_t));
        l2->count++;
    }

    *pe = *page;
}

/*
Returnerer side informationen for den logiske side 'logical'.
Hvis den logiske side ikke eksisterer returneres 'NULL'.
*/
page_entry_t *pagetable_get(uint32_t logical) {
    level_t *l2;
    uint32_t p1, p2;

    p1 = (logical >> LEVEL1_BIT) & LEVEL_MASK;

    l2 = level1.childs[p1];

    if (l2 != NULL) {
        p2 = (logical >> LEVEL2_BIT) & LEVEL_MASK;
        return l2->childs[p2];
    }

    return NULL;
}
