#include <stdlib.h>
#include <inttypes.h>
#include <stdio.h>			// Inkluderes s vi kan bruge printf()

#include "tlb.h"
#include "mem.h"

#define TLB_SIZE 8

#define TLB_USED           0x1
#define TLB_REFERENCED     0x2

typedef struct {
    uint32_t logical;       // Logisk side
    uint32_t physical;      // Fysisk side
    uint16_t protection;    // Tilgangsrettigheder
    uint16_t flags;         // Flag, TLB_USED, TLB_REFERENCED
} tlb_entry_t;

static tlb_entry_t tlb[TLB_SIZE];
	
// Variabel, der bruges til at holde styr p second-chance algoritmen
int chance;	

/* Initialiser TLB'en. */
void tlb_init() {
    int i;

    for(i=0; i<TLB_SIZE; i++) {
        tlb[i].flags      = 0;
    }

	chance = 0;
}

/*
Indstter den logiske side 'logical' som referererer til den fysiske side
(frame) 'physical'. Parameteren 'protection' specificerer hvilke
tilgangsrettigheder der skal vre til siden (read og/eller write).
*/
void tlb_insert(uint32_t logical, uint32_t physical, uint16_t protection) {
    int i;

	/* Fjern offset. */
    logical &= PAGE_MASK;

    /* Led efter et tomt element. */
    for(i=0; i<TLB_SIZE; i++) {
        if (!(tlb[i].flags & TLB_USED)) {
            break;
        }
    }

    if (i == TLB_SIZE) {	// Der blev ikke fundet et tomt element.
	
        /* 
		Find en kandidat til udskiftning vha. second-chance algoritmen.
 		Algoritmen leder fra og med plads chance efter urefereret TLB entry
		og giver second chances til de allerede refererede entries, der 
		passeres.
		*/

		while(tlb[chance].flags & TLB_REFERENCED) {
			tlb[chance].flags = TLB_USED;
			chance = (chance + 1) % TLB_SIZE;	// Flyt pointeren til nste TLB entry
		}
		
		i = chance;
		printf("Offer fundet p plads %d\n", i);
		chance = (chance + 1) % TLB_SIZE;		// Flyt pointeren til nste TLB entry
    }

    tlb[i].logical    = logical;
    tlb[i].physical   = physical;
    tlb[i].protection = protection;
    tlb[i].flags      = TLB_USED | TLB_REFERENCED;

	printf("Indsat TLB entry, log: 0x%x, phys: 0x%x p plads %d\n", logical, physical, i);
}

/*
Laver et opslag p den logiske adresse 'logical'. Den tilsvarende
fysiske side gemmes i den variabel som 'physical' peger p. Funktionen
returnerer tilgangsrettighederne.

Hvis siden for den logiske adresse ikke findes i TLB'en kaldes
funktionen 'tlb_fault'.
*/
uint16_t tlb_lookup(uint32_t logical, uint32_t *physical) {
    int i;

    /* Fjern offset. */
    logical &= PAGE_MASK;

    do {
        /* Find siden */
        for(i=0; i<TLB_SIZE; i++) {
            if (tlb[i].logical == logical) {
                tlb[i].flags |= TLB_REFERENCED;
                *physical = tlb[i].physical;
                return tlb[i].protection;
            }
        }
    } while (tlb_fault(logical));
    return 0;
}
